import socket
import ssl
import sys
import task_information
import time
import json
import base64
import threading
import bcrypt
import hashlib
import secrets
import client_crypto

memory_lock = threading.RLock()

_version = "2.0"
def get_version():
    return _version

# --- not recommended to use outside this file --- #
default_server_socket: ssl.SSLSocket = None

# this function will initialize a server_socket.
# After calling this function you don't need to tell the other functions the server_socket anymore. But you still can.
def initialize_server_socket(server_socket: ssl.SSLSocket):
    global default_server_socket
    default_server_socket = server_socket

# --- not recommended to get called outside this file --- #
def recv_msg_as_bytes(server_socket: ssl.SSLSocket = None) -> bytes:
    if server_socket == None: server_socket = default_server_socket

    """
    Every recv in this code should formally be an recv_exact,
    but python doesn't implement one by default.
    However it also works with a normal recv.
    """

    CHUNK_SIZE = 4096

    data = server_socket.recv(4)

    # Casting UTF8/ASCII-formatted String to int.
    number_of_bytes = None
    for i in range(len(data)):
        if data[i] == 0:
            """
            By entering this if-statement we know that our message ends at index i.
            So we read everything from data until i -> list(data)[:i]. (Casting to list is needed, because we cant use a bytes object like this)
            Then we cast it firstly to a string.
            This String represents the number of bytes we want to receive in the following part. So we cast it to an Integer and assign it to number_of_bytes.
            """
            number_of_bytes = int(
                str(
                    bytes(
                        list(data)[:i]
                    ),
                    'utf8'
                )
            )
            break  # The end of the msg is reached.
    if number_of_bytes is None:  # in that case all bytes are used, so there is no zero byte
        number_of_bytes = int(str(data, 'utf8'))
    
    # Special Case: number_of_bytes == 0: It means that 4 bytes aren't enough to send the length of the message
    if number_of_bytes == 0:
        data = server_socket.recv(16)
        number_of_bytes = None
        for i in range(len(data)):
            if data[i] == 0:
                number_of_bytes = int(
                    str(
                        bytes(
                            list(data)[:i]
                        ),
                            'utf8'
                    )
                )
                break
        if number_of_bytes is None: # in that case all bytes are used, so there is no zero byte
            number_of_bytes = int(str(data, 'utf8'))
    
    if number_of_bytes > CHUNK_SIZE:
        number_of_chunks = number_of_bytes // CHUNK_SIZE
        last_chunk_size = number_of_bytes % CHUNK_SIZE
        data = b"" # empty byte array
        for i in range(number_of_chunks):
            data += server_socket.recv(CHUNK_SIZE)
        data += server_socket.recv(last_chunk_size)
    else:
        data = server_socket.recv(number_of_bytes)
    
    return data

# --- not recommended to get called outside this file --- #
def recv_msg(server_socket: ssl.SSLSocket = None) -> str:
    if server_socket == None: server_socket = default_server_socket

    return str(recv_msg_as_bytes(server_socket=server_socket), 'utf8')

# --- not recommended to get called outside this file --- #
def recv_msg_as_list(server_socket: ssl.SSLSocket = None) -> list:
    if server_socket == None: server_socket = default_server_socket

    return list(recv_msg_as_bytes(server_socket=server_socket))

# --- not recommended to get called outside this file --- #
def recv_msg_as_json(server_socket: ssl.SSLSocket = None) -> dict:
    if server_socket == None: server_socket = default_server_socket

    return json.loads(recv_msg(server_socket=server_socket))

# --- not recommended to get called outside this file --- #
def send_msg_as_bytes(msg: bytes, server_socket: ssl.SSLSocket = None):
    if server_socket == None: server_socket = default_server_socket

    CHUNK_SIZE = 4096
    
    length = len(msg)
    length_str = str(length)
    length_str_length = len(length_str)

    """
    Normally the server listens for exactly 4 bytes, which represent the length of the following message.
    If 4 bytes aren't enough, send "0" as UTF-8/ASCII encoded String to let the server listen for 16 bytes.
    Mor than 16 bytes are out of convention. (It would be many PB)
    """
    if length_str_length <= 4:
        """
        Sending a buffer, which content is the length of the following message in UTF-8/ASCII.
        It is important that the UTF-8 encoded String ends with a zero byte or fills all 4 bytes.
        """
        buff = [0,0,0,0]
        for i in range(len(length_str)):
            buff[i] = bytes(length_str, 'utf8')[i]
        server_socket.sendall(bytes(buff))

    elif length_str_length <= 16:
        server_socket.sendall(bytes([48, 0, 0, 0])) # 48 is ASCII/UTF-8 for "0"
        
        buff = [0, 0, 0, 0,
                0, 0, 0, 0,
                0, 0, 0, 0,
                0, 0, 0, 0]
        for i in range(len(length_str)):
            buff[i] = bytes(length_str, 'utf8')[i]
        
        server_socket.sendall(bytes(buff))
    else:
        pass # undefined

    """
    If the length of the message is longer than CHUNK_SIZE,
    it has to be sent in different chunks.
    Oterwhise it can just be sent in once.
    """
    if length > CHUNK_SIZE:
        number_of_chunks = length // CHUNK_SIZE

        for i in range(number_of_chunks):
            server_socket.sendall(
                msg[(CHUNK_SIZE * i):(CHUNK_SIZE * (i + 1))])
        server_socket.sendall(msg[number_of_chunks * CHUNK_SIZE:])
    else:
        server_socket.sendall(msg)


# --- not recommended to get called outside this file --- #
def send_msg(msg: str, server_socket: ssl.SSLSocket = None):
    if server_socket == None: server_socket = default_server_socket

    b_msg = bytes(msg, 'utf8')
    send_msg_as_bytes(server_socket=server_socket, msg=b_msg)

# --- not recommended to get called outside this file --- #
def send_msg_as_list(msg: list, server_socket: ssl.SSLSocket = None):
    if server_socket == None: server_socket = default_server_socket

    l_msg = bytes(msg)
    send_msg_as_bytes(server_socket=server_socket, msg=l_msg)

# --- not recommended to get called outside this file --- #
def send_msg_as_json(msg: dict, indent = None, sort_keys=False, server_socket: ssl.SSLSocket = None):
    if server_socket == None: server_socket = default_server_socket

    json_msg = json.dumps(msg, indent=indent, sort_keys=sort_keys)
    send_msg(msg=json_msg, server_socket=server_socket)

# --- not recommended to get called outside this file --- #
def send_task(task: list, server_socket: ssl.SSLSocket = None):
    if server_socket == None: server_socket = default_server_socket

    json_dict = {"t": task[0]}
    if len(task) > 1:
        json_dict.update({"p": task[1:]})

    send_msg_as_json(server_socket=server_socket, msg=json_dict)

"""
Checking Versions between Client and Server for compatibility.
The server returns a JSON-String as following:
"v": version_of_server -> str (for example "1.5.7")
"r": result -> int (0 means, that the version of the client is supported. Every other means it is not)
"""
def start_task_api_version_check(server_socket: ssl.SSLSocket = None) -> dict:
    if server_socket == None: server_socket = default_server_socket

    json_dict = {"v": get_version()} # sending the current version of the client

    send_msg_as_json(msg=json_dict, server_socket=server_socket)
    ret_val = recv_msg_as_json(server_socket=server_socket)
    return ret_val

"""
this function will return an integer, which tells you the outcome of the task:
-1   In login queue. Another result will follow soon.
0    Successfull. You are now logged in.
1    Undefined Error. If this happens, it is probably the fault of the server.
2    Already logged in as another user.
3    Sended information is invalid (for example there is something missing).
4    Wrong password
5    User doesn't exist
"""
def start_task_login(username: str, password: str, task_parameters: list = task_information.login.default_task_parameters, server_socket: ssl.SSLSocket = None) -> dict:
    if server_socket == None: server_socket = default_server_socket

    # starting the task
    task_list: list = [task_information.login.task_number] + task_parameters
    send_task(task=task_list, server_socket=server_socket)

    """
    To hash our passwor, we first need the pwd_salt, which is saved on the server.
    So we send the username to the server and receive the pwd_salt from it.
    """
    information_dict = {"u": username}
    send_msg_as_json(msg=information_dict, server_socket=server_socket)

    result = recv_msg_as_json(server_socket=server_socket)
    if result["r"] != 0:
        return result
    
    b64_pwd_salt = result["s"]
    pwd_salt = base64.b64decode(b64_pwd_salt)

    # hashing password
    pwd_bytes = bytes(password, 'utf8')
    scrypt_pwd, _ = client_crypto.SCRYPT.hashpw(pwd_bytes, pwd_salt)
    b64pwd: str = str(base64.b64encode(scrypt_pwd), 'utf8')

    # Sending the hashed and base64 encoded password to the server.
    information_dict = {"p": b64pwd}
    send_msg_as_json(msg=information_dict, server_socket=server_socket)

    # Handling the login queue
    while True:
        task_result: dict = recv_msg_as_json(server_socket=server_socket)
        if task_result["r"] == -1:
            _queue_position = task_result["p"]  # can be displayed to the user
            continue
        else:
            break

    # return the outcome of the task
    return task_result
    
"""
this function will return an integer, which tells you the outcome of the task:
0    successfull. You are now logged in.
1    Undefined Error. If this happens, it is probably the fault of the server.
2    Something went wrong. The username is probably already given or not usable for some reason.
3    Sended information is invalid (for example there is something missing).
4    Username already in use or invalid
5    Requirements for password, key etc. not fullfilled
        - pwd has to be exactly 64 bytes long (This is the length of the scrypt hash)
        - pwd_salt has to be exactly 16 bytes long
        - key has to be exactly 32 bytes long
        - key salt has to be exactly 32 bytes long
"""
def start_task_register(username: str, password: str,
 task_parameters: list = task_information.register.default_task_parameters, server_socket: ssl.SSLSocket = None) -> dict:
    if server_socket == None: server_socket = default_server_socket

    # starting the task
    task_list: list = [task_information.register.task_number] + task_parameters
    send_task(task=task_list, server_socket=server_socket)

    # hashing password
    pwd_bytes = bytes(password, 'utf8')
    scrypt_pwd, scrypt_pwd_salt = client_crypto.SCRYPT.hashpw(pwd_bytes)
    b64pwd: str = str(base64.b64encode(scrypt_pwd), 'utf8')
    b64pwd_salt: str = str(base64.b64encode(scrypt_pwd_salt), 'utf8')

    # Creation of the standard AES key of the client
    key: bytes = secrets.token_bytes(32) # creating key of 32 bytes (256 Bit) with a cryptographically secure pseudorandom function
    encrypted_key, encrypted_key_salt = client_crypto.AES.encrypt(key, pwd_bytes)
    b64encrypted_key: str = str(base64.b64encode(encrypted_key), 'utf8')
    b64encrypted_key_salt: str = str(base64.b64encode(encrypted_key_salt), 'utf8')

    information_dict = {"u": username,
                        "p": b64pwd,
                        "i": b64pwd_salt,
                        "k": b64encrypted_key,
                        "s": b64encrypted_key_salt}
    send_msg_as_json(msg=information_dict, server_socket=server_socket)

    # return the outcome of the task
    task_result: dict = recv_msg_as_json(server_socket=server_socket)
    return task_result

"""
this function will return an integer, which tells you the outcome of the task:
0    successfull. You are no longer logged in.
"""
def start_task_logout(task_parameters: list = task_information.logout.default_task_parameters, server_socket: ssl.SSLSocket = None) -> dict:
    if server_socket == None: server_socket = default_server_socket

    # starting the task
    task_list: list = [task_information.logout.task_number] + task_parameters
    send_task(task=task_list, server_socket=server_socket)

    # return the outcome of the task
    task_result: dict = recv_msg_as_json(server_socket=server_socket)
    return task_result

"""
this function will return an integer, which tells you the outcome of the task:
0     Everything was successfull
1     Undefined Error
2     Client isn't logged in
3     Label invalid
"""
def start_task_memory(key: bytes = None,
                    label: str = None, enc_label: str = None,
                    msg: str = None,
                    receiver: str = "self", repeat: bool = False,
                    start_point: int = None, stop_point: int = None,
                    frequence: int = None, time_point = None,
                    since: int = None,
                    task_parameters: list = task_information.memory.default_task_parameters, server_socket: ssl.SSLSocket = None) -> dict:
    global memory_lock
    if label is not None:
        encrypted_label, encrypted_label_salt = client_crypto.AES.encrypt(bytes(label, 'utf8'), key)
        b64_encrypted_label: str = str(base64.b64encode(encrypted_label), 'utf8')
        b64_encrypted_label_salt: str = str(base64.b64encode(encrypted_label_salt), 'utf8')
        label = b64_encrypted_label
        label_salt = b64_encrypted_label_salt
    with memory_lock:
        if server_socket == None: server_socket = default_server_socket

        # starting the task
        task_list: list = [task_information.memory.task_number] + task_parameters
        send_task(server_socket=server_socket, task=task_list)

        information_dict = {}
        # evaluate the task_parameters
        if task_parameters[0] == 0:
            if task_parameters[1] == 0:
                if len(task_parameters) > 2 and task_parameters[2] == 1:
                    pass # nothing to do here
                else:
                    information_dict.update({"s": since})
            elif task_parameters[1] == 1:
                information_dict.update({"l": enc_label})
        elif task_parameters[0] == 1: # We want to send a new memory task to the server
            # sending all information in the right order
            information_dict.update({"l": label})
            information_dict.update({"ls": label_salt})
            memory_information = {}
            memory_information.update({"m": msg})
            memory_information.update({"r": repeat})

            # We have to send different information based on repeat
            if repeat:
                memory_information.update({"s1": start_point})
                memory_information.update({"s2": stop_point})
                memory_information.update({"f": frequence})
            else:
                memory_information.update({"t": time_point})
            encrypted_memory_information, encrypted_memory_information_salt = client_crypto.AES.encrypt(bytes(json.dumps(memory_information), 'utf8'), key)
            information_dict.update({"i": str(base64.b64encode(encrypted_memory_information), 'utf8')})
            information_dict.update({"s": str(base64.b64encode(encrypted_memory_information_salt), 'utf8')})
        elif task_parameters[0] == 2: # editing a task
            information_dict.update({"l": enc_label})
            memory_information = {}
            memory_information.update({"m": msg})
            memory_information.update({"r": repeat})

            # We have to send different information based on repeat
            if repeat:
                memory_information.update({"s1": start_point})
                memory_information.update({"s2": stop_point})
                memory_information.update({"f": frequence})
            else:
                memory_information.update({"t": time_point})
            encrypted_memory_information, encrypted_memory_information_salt = client_crypto.AES.encrypt(bytes(json.dumps(memory_information), 'utf8'), key)
            information_dict.update({"i": str(base64.b64encode(encrypted_memory_information), 'utf8')})
            information_dict.update({"s": str(base64.b64encode(encrypted_memory_information_salt), 'utf8')})
        elif task_parameters[0] == 3: # deleting a task
            information_dict.update({"l": enc_label})

    
        send_msg_as_json(msg=information_dict, server_socket=server_socket)

        # return the outcome of the task
        task_result: dict = recv_msg_as_json(server_socket=server_socket)
        return task_result

"""
WARNING! THIS FEATURE ISN'T SUPPORTED YET!

the will return the outcome of the operation in json format (outcome is "result")
0     Everything was successfull
1     Unexpected Error (The token is probably still there)
2     Token invalid
3     Wrong format. Error with JSON.
4     User not logged in/already logged in
"""
def start_task_login_token(username: str = None, token: str = None, old_token: str = None, task_parameters: list = task_information.login_token.default_task_parameters, server_socket: ssl.SSLSocket = None) -> dict:
    if server_socket == None: server_socket = default_server_socket

    # starting the task
    task_list: list = [task_information.login_token.task_number] + task_parameters
    send_task(server_socket=server_socket, task=task_list)

    if task_parameters == task_information.login_token.login_task_parameters:
        information_dict: dict = {"u": username, "t": token}
        send_msg_as_json(information_dict, server_socket=server_socket)
    elif task_parameters == task_information.login_token.get_new_token_task_parameters:
        # returns "result", "token", "start_point" and "stop_point"
        information_dict: dict = {"o": old_token}
        send_msg_as_json(information_dict, server_socket=server_socket)
    elif task_parameters == task_information.login_token.delete_token_task_parameters:
        information_dict: dict = {"t": token}
        send_msg_as_json(information_dict, server_socket=server_socket)
    elif task_parameters == task_information.login_token.delete_all_tokens_task_parameters:
        pass # nothing has to be done here

    # return the outcome of the task
    task_result: dict = recv_msg_as_json(server_socket=server_socket)
    return task_result
